//############################################################################
//##                                                                        ##
//##  FONT.CPP: System font acquisition and management for Win32 OpenGL     ##
//##                                                                        ##
//##  Version 1.00 of 6-Aug-99: Initial                                     ##
//##                                                                        ##
//##  Project: Eden runtime interpreter                                     ##
//##   Author: John Miles                                                   ##
//##                                                                        ##
//##  C++ source compatible with MSVC 12                                    ##
//##                                                                        ##
//############################################################################
//##                                                                        ##
//## Released to the public domain 12/15/99 by John Miles (jmiles@pop.net)  ##
//## No restrictions are placed on use or distribution of any of this code. ##
//##                                                                        ##
//############################################################################

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <windowsx.h>
#include <winuser.h>
#include <mmsystem.h>
#include <gl/gl.h>

#include <stdio.h>
#include <conio.h>
#include <dos.h>
#include <malloc.h>
#include <direct.h>
#include <io.h>
#include <time.h>
#include <assert.h>
#include <math.h>
#include <limits.h>

#include "system.h"
#include "userio.h"
#include "stdtpl.h"

//
// [TODO: support other font formats besides 32bpp]
// [TODO: support drop shadows]
//

//
// Configuration equates
//

#define N_CHARS    256     // We only support single-byte charsets for now

#define MAX_TXT_W  256     // Maximum texture width and height
#define MAX_TXT_H  256                      

//
// Font structure
//

struct GLFONT
{
   GLFONT *next;          
   GLFONT *prev;       
                             
   S32    index;                     // Index of this entry in linear array

   void   initialize(void *V) {}     // Store indicated object in T
   void   shutdown  (void)    {}     // Called when unlinked from pool
   void   display   (void)    {}     // Optionally display contents of T

   //
   // User vars
   //

   GLVTABLE *GL;

   int char_width [N_CHARS];        
   S32 char_x     [N_CHARS];
   S32 char_y     [N_CHARS];

   S8 *kern_l     [N_CHARS];         // First non-background pixel in each row of character cell (-1 if row empty)
   S8 *kern_r     [N_CHARS];         // Last non-background pixel in each row of character cell
   S32 min_l      [N_CHARS];         // Leftmost kern_l value across all character rows
   S32 max_r      [N_CHARS];         // Rightmost kern_r value across all character rows

   S32 cell_w;                       // Maximum character X,Y size in
   S32 cell_h;                       // pixels
   S32 cells_per_texture;            // # of character cells per texture

   S32 txt_w;                        // Size of each texture in pixels
   S32 txt_h;

   GLuint *textures;                 // GL texture-handle array
   S32     n_textures;

   S32 char_spacing;                 // # of pixels to leave between adjacent chars

   FNTFLAGS flags;                   // Flags
};

//
// Globals
// 

Pool <GLFONT, 16> fonts;             // List of managed fonts
                                     
/*********************************************************************/
//
// create_font()  
//
/*********************************************************************/

HGLFONT  USERIO::create_font (GLVTABLE *GL, //)
                              HFONT     hFont,
                              COLORREF  crColors,
                              FNTFLAGS  creation_flags,
                              S32       explicit_character_spacing)
{
   HGLFONT fnt = fonts.allocate();

   GLFONT *F = &fonts.list[fnt];

   F->GL = GL;

   //
   // Create 32bpp DIB section
   //

   U32 *lpDIBBuffer = NULL;

   BITMAPINFO *pbmi = (BITMAPINFO *) 
          malloc(sizeof (BITMAPINFOHEADER) + (sizeof (RGBQUAD) * 256));

   memset(pbmi, 
          0, 
          sizeof (BITMAPINFOHEADER) + (sizeof (RGBQUAD) * 256));

   BITMAPINFOHEADER *pbmih = &(pbmi->bmiHeader);

   pbmih->biSize          =  sizeof(*pbmih);
   pbmih->biWidth         =  MAX_TXT_W;
   pbmih->biHeight        = -MAX_TXT_H;
   pbmih->biPlanes        =  1;
   pbmih->biBitCount      =  32;
   pbmih->biSizeImage     =  0;
   pbmih->biXPelsPerMeter =  0;
   pbmih->biYPelsPerMeter =  0;
   pbmih->biClrUsed       =  0;
   pbmih->biClrImportant  =  0;
   pbmih->biCompression   =  BI_BITFIELDS;

   *(U32 *) (&(pbmi->bmiColors[0])) = 0xff0000;
   *(U32 *) (&(pbmi->bmiColors[1])) = 0x00ff00;
   *(U32 *) (&(pbmi->bmiColors[2])) = 0x0000ff;

   HBITMAP hDIB = CreateDIBSection(NULL,            // Device context
                                   pbmi,            // BITMAPINFO structure
                                   DIB_RGB_COLORS,  // Color data type
                        (void **) &lpDIBBuffer,     // Address of image map pointer
                                   NULL,            // File
                                   0);              // Bitmap file offset
   assert(hDIB != NULL);

   //
   // Create display DC and map it to our DIB
   //

   HDC hDC = CreateCompatibleDC(0);
   assert(hDC != NULL);

   HBITMAP hDefaultBitmap = SelectBitmap(hDC, hDIB);

   //
   // Select specified Windows font into compatible DC
   //

   SelectObject(hDC, hFont);
   SetTextColor(hDC, crColors);

   //
   // Build lookup table for character widths
   //                                     

   TEXTMETRIC tm;

	GetTextMetrics(hDC, &tm);

   GetCharWidth(hDC,
                0,
                N_CHARS-1,
                F->char_width);

   //
   // If font color is not black, set background color to black 
   // (so we can sample white text properly)
   //
   // If font is black, we leave the background color alone for contrast
   //

   if (crColors != 0)
      {
      SetBkColor(hDC, RGB(0,0,0));
      }

   //
   // Get spacing for character cells in bitmap
   //

   F->txt_w = MAX_TXT_W;
   F->txt_h = MAX_TXT_H;

   F->cell_w = tm.tmMaxCharWidth;
   F->cell_h = tm.tmHeight;

   if ((F->cell_w > MAX_TXT_W) || (F->cell_h > MAX_TXT_H))
      {
      debug_printf("USERIO: Font too large!\n");

      DeleteDC(hDC);
      fonts.unlink(fnt);
      return -1;
      }

   //
   // Allocate and initialize kerning arrays
   //

   S32 i;

   for (i=0; i < N_CHARS; i++)
      {
      F->kern_l[i] = (S8 *) malloc(tm.tmHeight);
      F->kern_r[i] = (S8 *) malloc(tm.tmHeight);

      assert(F->kern_l[i]);
      assert(F->kern_r[i]);

      memset(F->kern_l[i], -1, tm.tmHeight);
      memset(F->kern_r[i],  0, tm.tmHeight);
      }

   //
   // Clear bitmap object to reserved transparent color which will
   // never appear in a valid font (i.e., not a selectable font color)
   //

   U32 xp_bkgnd = 0xDEFACED;

   fill_U32(lpDIBBuffer,
            xp_bkgnd,
            MAX_TXT_W * MAX_TXT_H);

   //
   // Draw space character to sample the font's actual background color
   //
   // This color will become alpha-transparent in the texture bitmap
   //

   TextOut(hDC, 0, 0, " ", 1);

   U32 font_bkgnd = xp_bkgnd;

   for (S32 y=0; (y < F->cell_h) && (font_bkgnd == xp_bkgnd); y++)
      {
      U32 *d = &lpDIBBuffer[y * MAX_TXT_W];

      for (S32 x=0; x < F->cell_w; x++)
         {
         if (d[x] != xp_bkgnd)
            {
            font_bkgnd = d[x];
            break;
            }
         }
      }

   assert(font_bkgnd != xp_bkgnd);

   //
   // Get # of actual MAX_TXT_W * MAX_TXT_H textures necessary to store font
   //

   debug_printf("USERIO: %d x %d cell size\n",F->cell_w,F->cell_h);

   S32 x_cells_per_txt = MAX_TXT_W / F->cell_w; if (x_cells_per_txt == 0) x_cells_per_txt = 1;
   S32 y_cells_per_txt = MAX_TXT_H / F->cell_h; if (y_cells_per_txt == 0) y_cells_per_txt = 1;

   F->cells_per_texture = x_cells_per_txt * y_cells_per_txt;

   debug_printf("USERIO: %d cells per %dx%d texture\n",F->cells_per_texture,MAX_TXT_W,MAX_TXT_H);

   F->n_textures = N_CHARS / F->cells_per_texture; if (F->n_textures == 0) F->n_textures = 1;

   debug_printf("USERIO: %d texture(s) needed\n",F->n_textures);

   //
   // Allocate required texture handles from OpenGL
   //

   F->textures = (GLuint *) malloc(F->n_textures * sizeof(GLuint));

   GL->glGenTextures(F->n_textures, F->textures);

   //
   // For each destination texture...
   //

   S32 ch = 0;

   for (S32 t=0; t < F->n_textures; t++)
      {
      //
      // Clear temporary texture buffer
      //

      fill_U32(lpDIBBuffer,
               xp_bkgnd,
               MAX_TXT_W * MAX_TXT_H);

      S32 cx = 0;
      S32 cy = 0;

      //
      // For each cell in destination texture...
      //

      for (S32 i=0; i < F->cells_per_texture; i++)
         {
         //
         // Render character into sampling bitmap
         //

         C8 string[2];

         string[0] = (C8) ch;
         string[1] = 0;

         TextOut(hDC,
                 cx,
                 cy,
                 string,
                 1);

         //
         // Scan top row to find left/right boundaries of character cell
         //
         // We assume the non-transparent cell is rectangular, with 
         // straight left/right edges, and occupies the entire vertical
         // cell space
         //

         S32 l = -1;
         S32 r = -1;

         U32 *d = &lpDIBBuffer[(cy * MAX_TXT_W) + cx];

         for (r=0; r < F->cell_w; r++)
            {
            if (d[r] == xp_bkgnd)
               {
               if (l != -1) break;
               }
            else
               {
               if (l == -1) l = r;
               }
            }
         --r;

         //
         // Sample character and copy it to temporary texture buffer
         //

         S8 *kl = F->kern_l[ch];
         S8 *kr = F->kern_r[ch];

         F->min_l[ch] = 127;
         F->max_r[ch] = -128;

         for (S32 y=0; y < F->cell_h; y++)
            {
            U32 *d = &lpDIBBuffer[((cy + y) * MAX_TXT_W) + cx + l];

            //
            // Locate kerning boundaries in this pixel row, and mark
            // character backgrounds alpha-transparent
            //
            // If this is a space character, its visible width at each row is 
            // considered to be the non-transparent cell width
            //

            S32 _kl = l;
            S32 _kr = r;

            if (ch != ' ')
               {
               _kl = -1;
               _kr = -1;

               for (S32 x=l; x <= r; x++,d++)
                  {
                  if (*d != font_bkgnd)
                     {
                     _kr = x;

                     if (_kl == -1)
                        {
                        _kl = x;
                        }

                     *d |= 0xff000000;
                     }
                  }
               }

            //
            // Advance kerning arrays to next pixel row
            //

            *kl++ = S8(_kl);
            *kr++ = S8(_kr);

            //
            // Keep track of total width of visible (non-background) 
            // portion of character
            //

            if ((_kl < F->min_l[ch]) && (_kl != -1))
               {
               F->min_l[ch] = _kl;
               }

            if (_kr > F->max_r[ch])
               {
               F->max_r[ch] = _kr;
               }
            }

         //
         // Advance cell destination coords
         //

         F->char_x[ch] = cx;
         F->char_y[ch] = cy;

         cx += F->cell_w;

         if ((cx + F->cell_w) > MAX_TXT_W)
            {
            cx = 0;
            cy += F->cell_h;
            }

         //
         // Exit loop if end of character set reached
         //

         if (++ch == N_CHARS)
            {
            break;
            }
         }

      //
      // Send texture to OpenGL
      //

      GL->glBindTexture(GL_TEXTURE_2D, F->textures[t]);

      GL->glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

      GL->glTexImage2D(GL_TEXTURE_2D, 
                       0, 
                       4,
                       MAX_TXT_W,
                       MAX_TXT_H,
                       0, 
                       GL_RGBA,
                       GL_UNSIGNED_BYTE,
                       lpDIBBuffer);

      GL->glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S,     GL_REPEAT);
      GL->glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T,     GL_REPEAT);
      GL->glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
      GL->glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
      }

   //
   // Delete temporary objects
   //

   SelectBitmap(hDC, hDefaultBitmap);
   DeleteDC(hDC);

   DeleteObject(hDIB);
   free(pbmi);

   //
   // Set default character spacing to (width of space character / 3)-1 
   // unless an explicit value is provided
   //

   if (explicit_character_spacing != -1)
      {
      F->char_spacing = explicit_character_spacing;
      }
   else
      {
      F->char_spacing = max(1, (F->char_width[' '] / 3) - 1);
      }

   //
   // If font's space character takes up an entire character cell, assume
   // this is a fixed-width font
   // 
   // Kerning on proportional fonts is enabled by default
   // 

   if (F->char_width[' '] == F->cell_w)
      {
      F->flags = creation_flags;
      }
   else
      {
      F->flags = creation_flags | FNT_IS_PROPORTIONAL | FNT_KERNING_ENABLED;
      }

   return fnt;
}

/*********************************************************************/
//
// destroy_font()  
//
/*********************************************************************/

void USERIO::destroy_font    (HGLFONT   font)
{
   if (font == -1)
      {
      return;
      }

   GLFONT *F = &fonts.list[font];

   F->GL->glDeleteTextures(F->n_textures, F->textures);

   fonts.unlink(font);
}

/*********************************************************************/
//
// draw_character()
//
/*********************************************************************/

S32  USERIO::draw_character  (HGLFONT   font, //)
                              S32       character,
                              S32       x,
                              S32       y)
{
   if (font == -1)
      {
      return 0;
      }

   GLFONT *F = &fonts.list[font];

   //
   // Get coordinates of character cell within texture, centered within
   // texel cells
   //
   // OpenGL skips the right and bottom edges, so these must be specified
   // with an offset of +1
   //

   SINGLE fl = SINGLE(F->char_x[character] + F->min_l[character]    ); 
   SINGLE ft = SINGLE(F->char_y[character]                          ); 
   SINGLE fr = SINGLE(F->char_x[character] + F->max_r[character] + 1); 
   SINGLE fb = SINGLE(F->char_y[character] + F->cell_h              ); 

   //
   // Get screen coordinates, centered within pixel cells
   //
   // We want a 1:1 mapping from texels to pixels
   //

   SINGLE sl = SINGLE(x);
   SINGLE st = SINGLE(y);
   SINGLE sr = sl + (fr - fl);
   SINGLE sb = st + (fb - ft);

   //
   // Normalize texture coordinates
   //

   fl /= SINGLE(F->txt_w);
   ft /= SINGLE(F->txt_h);
   fr /= SINGLE(F->txt_w);
   fb /= SINGLE(F->txt_h);

   //
   // Bind character's source texture
   //

   S32 t = character / F->cells_per_texture;

   F->GL->glBindTexture(GL_TEXTURE_2D, F->textures[t]);

   //
   // Render the character with standard texture alpha blending
   //

   F->GL->glEnable(GL_TEXTURE_2D);

   F->GL->glEnable(GL_BLEND);
   F->GL->glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

   F->GL->glBegin(GL_QUADS);

   F->GL->glTexCoord2f(fl,ft); F->GL->glColor3f(1.0F, 1.0F, 1.0F); F->GL->glVertex2f(sl,st);
   F->GL->glTexCoord2f(fr,ft); F->GL->glColor3f(1.0F, 1.0F, 1.0F); F->GL->glVertex2f(sr,st);
   F->GL->glTexCoord2f(fr,fb); F->GL->glColor3f(1.0F, 1.0F, 1.0F); F->GL->glVertex2f(sr,sb);
   F->GL->glTexCoord2f(fl,fb); F->GL->glColor3f(1.0F, 1.0F, 1.0F); F->GL->glVertex2f(sl,sb);

   F->GL->glEnd();

   //
   // Return width of visible portion of character
   //

   return S32(fr - fl) + 1;
}

/*********************************************************************/
//
// draw_string()
//
/*********************************************************************/

S32 USERIO::draw_string       (HGLFONT   font, //)
                               C8       *string,
                               S32       x,
                               S32       y)
{
   S32 prev_x    = x;
   S32 prev_char = -1;
   S32 w = 0;

   while (1)
      {
      C8 c = *string++;

      if (!c)
         {
         break;
         }

      S32 cx;

      kern_pair(font,
                prev_char,
                prev_x,
                c,
               &cx);

      w = draw_character(font,
                         c,
                         cx,
                         y);

      prev_char = c;
      prev_x    = cx;
      }

   return prev_x + w;
}

/*********************************************************************/
//
// kern_pair()
//
/*********************************************************************/

void USERIO::kern_pair     (HGLFONT   font, //)
                            S32       prev_character,
                            S32       prev_x,
                            S32       new_character,
                            S32      *new_x)
{
   if (font == -1)
      {
      return;
      }

   GLFONT *F = &fonts.list[font];

   if (prev_character == -1)
      {
      *new_x = prev_x;
      return;
      }

   //
   // If this is a fixed-width font, or we're forcing fixed-width rendering,
   // characters are spaced by their cell size
   //

   if  ((F->flags & FNT_RENDER_AS_FIXED) || 
      (!(F->flags & (FNT_IS_PROPORTIONAL | FNT_RENDER_AS_PROP))))
      {
      *new_x = prev_x + F->cell_w;
      return;
      }

   //
   // Otherwise, get width of previous character's visible outline
   //
   // This is _not_ the same as the value returned from GetCharWidth() -- 
   // many characters aren't left-justified in their cells for some reason,
   // and/or include padding to the right.  We can obtain a much more 
   // consistent screen-font presentation by dynamically measuring each
   // character at the time the font is creaed.
   //

   S32 prev_w = (F->max_r[prev_character] - F->min_l[prev_character] + 1);

   //
   // In the absence of kerning, the X position for the new character is
   // determined by the position of the previous character plus its width
   //

   S32 x = prev_x + prev_w;

   if (F->flags & FNT_KERNING_ENABLED)
      {
      //
      // Try to move the characters closer together, by minimizing the 
      // distance between their visible rows 
      //

      S32 dist = LONG_MAX;

      S8 *nl = F->kern_l[new_character];    
      S8 *pr = F->kern_r[prev_character];

      for (S32 y=0; y < F->cell_h; y++)
         {
         S32 _pr = *pr++;
         S32 _nl = *nl++;

         if ((_pr == -1) || (_nl == -1))
            {
            continue;
            }

         S32 l = prev_x + (_pr - F->min_l[prev_character]);
         S32 r = x      + (_nl - F->min_l[new_character]);

         S32 w = r-l;

         if (w < dist)
            {
            dist = w;
            }
         }

      if (dist != LONG_MAX)
         {
         x -= (dist - 1);
         }
      }

   //
   // If neither new nor prev character is a space, add desired 
   // intercharacter spacing to new X position before returning it
   //

   if ((new_character == ' ') || (prev_character == ' '))
      {
      *new_x = x;
      }
   else
      {
      *new_x = x + F->char_spacing;
      }
}

/*********************************************************************/
//
// font_statistics()
//
/*********************************************************************/

FNTFLAGS USERIO::font_statistics   (HGLFONT   font, //)
                                    S32      *cell_size_X,
                                    S32      *cell_size_Y,
                                    S32      *space_width)
{
   if (font == -1)
      {
      return 0;
      }

   GLFONT *F = &fonts.list[font];
   
   if (cell_size_X != NULL) *cell_size_X = F->cell_w;
   if (cell_size_Y != NULL) *cell_size_Y = F->cell_h;
   if (space_width != NULL) *space_width = F->char_width[' '];

   return F->flags;
}
