//############################################################################
//##                                                                        ##
//##  FPMATH.C: Floating-point vector and matrix math routines              ##
//##                                                                        ##
//##  Flat-model source compatible with IBM 32-bit ANSI C/C++               ##
//##                                                                        ##
//##  Version 1.00 of 17-Sep-95: Initial                                    ##
//##                                                                        ##
//##  Author: John Miles                                                    ##
//##                                                                        ##
//############################################################################
//##                                                                        ##
//## Released to the public domain 12/15/99 by John Miles (jmiles@pop.net)  ##
//## No restrictions are placed on use or distribution of any of this code. ##
//##                                                                        ##
//############################################################################

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <math.h>

#include "typedefs.h"
#include "fpmath.h"

const TRANSFORM identity =
{   
   {1.0, 0.0, 0.0, 0.0, 
    0.0, 1.0, 0.0, 0.0, 
    0.0, 0.0, 1.0, 0.0}
};

//############################################################################
//##                                                                        ##
//## Multiplies source vector by rotation matrix, storing the result in     ##
//## dest                                                                   ##
//##                                                                        ##
//## (4x1 * 4x4 = 4x1 without translation)                                  ##
//##                                                                        ##
//## Assumes bottom row of rotation matrix is 0 0 0 1                       ##
//##                                                                        ##
//############################################################################

void VECTOR_rotate(VECTOR *src, VECTOR *dest, TRANSFORM *r)
{
   dest->x = (r->element[0][0] * src->x) +
             (r->element[0][1] * src->y) +
             (r->element[0][2] * src->z);

   dest->y = (r->element[1][0] * src->x) +
             (r->element[1][1] * src->y) +
             (r->element[1][2] * src->z);

   dest->z = (r->element[2][0] * src->x) +
             (r->element[2][1] * src->y) +
             (r->element[2][2] * src->z);
}

//############################################################################
//##                                                                        ##
//## Multiplies source vector by rotation matrix, storing the (translated)  ##
//## result in dest                                                         ##
//##                                                                        ##
//## (4x1 * 4x4 = 4x1)                                                      ##
//##                                                                        ##
//## Assumes bottom row of rotation matrix is 0 0 0 1                       ##
//##                                                                        ##
//############################################################################

void VECTOR_rotate_translate(VECTOR *src, VECTOR *dest, TRANSFORM *r)
{
   dest->x = (r->element[0][0] * src->x) +
             (r->element[0][1] * src->y) +
             (r->element[0][2] * src->z) +
             (r->element[0][3]);

   dest->y = (r->element[1][0] * src->x) +
             (r->element[1][1] * src->y) +
             (r->element[1][2] * src->z) +
             (r->element[1][3]);

   dest->z = (r->element[2][0] * src->x) +
             (r->element[2][1] * src->y) +
             (r->element[2][2] * src->z) +
             (r->element[2][3]);
}

//############################################################################
//##                                                                        ##
//## Multiplies source vector by inverted rotation matrix,                  ##
//## storing the result in dest                                             ##
//##                                                                        ##
//## (4x1 * 4x4 = 4x1 without translation)                                  ##
//##                                                                        ##
//## Assumes bottom row of rotation matrix is 0 0 0 1                       ##
//##                                                                        ##
//############################################################################

void VECTOR_inverse_rotate(VECTOR *src, VECTOR *dest, TRANSFORM *r)
{
   dest->x = (r->element[0][0] * src->x) +
             (r->element[1][0] * src->y) +
             (r->element[2][0] * src->z);

   dest->y = (r->element[0][1] * src->x) +
             (r->element[1][1] * src->y) +
             (r->element[2][1] * src->z);

   dest->z = (r->element[0][2] * src->x) +
             (r->element[1][2] * src->y) +
             (r->element[2][2] * src->z);
}

//############################################################################
//##                                                                        ##
//## Multiplies translated source vector by inverted rotation matrix,       ##
//## storing the result in dest                                             ##
//##                                                                        ##
//## (4x1 * 4x4 = 4x1)                                                      ##
//##                                                                        ##
//## Assumes bottom row of rotation matrix is 0 0 0 1                       ##
//##                                                                        ##
//############################################################################

void VECTOR_inverse_rotate_translate(VECTOR *src, VECTOR *dest, TRANSFORM *r)
{
   SINGLE x,y,z;

   //
   // Invert translation of source vector
   //

   x = src->x - r->element[0][3];
   y = src->x - r->element[1][3];
   z = src->x - r->element[2][3];

   //
   // Multiply inverse-translated source vector by inverted rotation transform
   //

   dest->x = (r->element[0][0] * x) +
             (r->element[1][0] * y) +
             (r->element[2][0] * z);

   dest->y = (r->element[0][1] * x) +
             (r->element[1][1] * y) +
             (r->element[2][1] * z);

   dest->z = (r->element[0][2] * x) +
             (r->element[1][2] * y) +
             (r->element[2][2] * z);
}

//############################################################################
//##                                                                        ##
//## Compare two vectors against given tolerance, returning 0 if equality   ##
//## criterion is met                                                       ##
//##                                                                        ##
//############################################################################

S32 VECTOR_compare(VECTOR *A, VECTOR *B, SINGLE tolerance)
{
   return ((fabs(A->x - B->x) > tolerance) ||
           (fabs(A->y - B->y) > tolerance) ||
           (fabs(A->z - B->z) > tolerance));
}

//############################################################################
//##                                                                        ##
//## Return exact Pythagorean magnitude of vector V                         ##
//##                                                                        ##
//############################################################################

SINGLE VECTOR_magnitude(VECTOR *V)
{
   return (SINGLE) sqrt((DOUBLE) (V->x * V->x +
                                  V->y * V->y +
                                  V->z * V->z));
}

//############################################################################
//##                                                                        ##
//## Return square magnitude of vector V (use when sqrt() unnecessary)      ##
//##                                                                        ##
//############################################################################

SINGLE VECTOR_square_magnitude(VECTOR *V)
{
   return (V->x * V->x +
           V->y * V->y +
           V->z * V->z);

}

//############################################################################
//##                                                                        ##
//## Return exact Pythagorean displacement of vectors V and W               ##
//##                                                                        ##
//############################################################################

SINGLE VECTOR_displacement(VECTOR *V, VECTOR *W)
{
   SINGLE dx,dy,dz;

   dx = V->x - W->x;
   dy = V->y - W->y;
   dz = V->z - W->z;

   return (SINGLE) sqrt(dx * dx +
                        dy * dy +
                        dz * dz);
}

//############################################################################
//##                                                                        ##
//## Return square displacement of vectors V and W (for comparison only)    ##
//##                                                                        ##
//############################################################################

SINGLE VECTOR_square_displacement(VECTOR *V, VECTOR *W)
{
   SINGLE dx,dy,dz;

   dx = V->x - W->x;
   dy = V->y - W->y;
   dz = V->z - W->z;

   return dx * dx +
          dy * dy +
          dz * dz;
}

//############################################################################
//##                                                                        ##
//## Multiplies matrix *M1 by matrix *M2, storing the result at *out        ##
//##                                                                        ##
//## (4x4 * 4x4 = 4x4)                                                      ##
//##                                                                        ##
//## Assumes bottom row of all matrices is 0 0 0 1                          ##
//##                                                                        ##
//############################################################################

void TRANSFORM_multiply(TRANSFORM *M1, TRANSFORM *M2, TRANSFORM *out)
{
      out->element[0][0] = M1->element[0][0] * M2->element[0][0] +
                           M1->element[0][1] * M2->element[1][0] +
                           M1->element[0][2] * M2->element[2][0];

      out->element[0][1] = M1->element[0][0] * M2->element[0][1] +
                           M1->element[0][1] * M2->element[1][1] +
                           M1->element[0][2] * M2->element[2][1];

      out->element[0][2] = M1->element[0][0] * M2->element[0][2] +
                           M1->element[0][1] * M2->element[1][2] +
                           M1->element[0][2] * M2->element[2][2];

      out->element[0][3] = M1->element[0][0] * M2->element[0][3] +
                           M1->element[0][1] * M2->element[1][3] +
                           M1->element[0][2] * M2->element[2][3] +
                           M1->element[0][3];

      out->element[1][0] = M1->element[1][0] * M2->element[0][0] +
                           M1->element[1][1] * M2->element[1][0] +
                           M1->element[1][2] * M2->element[2][0];

      out->element[1][1] = M1->element[1][0] * M2->element[0][1] +
                           M1->element[1][1] * M2->element[1][1] +
                           M1->element[1][2] * M2->element[2][1];

      out->element[1][2] = M1->element[1][0] * M2->element[0][2] +
                           M1->element[1][1] * M2->element[1][2] +
                           M1->element[1][2] * M2->element[2][2];

      out->element[1][3] = M1->element[1][0] * M2->element[0][3] +
                           M1->element[1][1] * M2->element[1][3] +
                           M1->element[1][2] * M2->element[2][3] +
                           M1->element[1][3];

      out->element[2][0] = M1->element[2][0] * M2->element[0][0] +
                           M1->element[2][1] * M2->element[1][0] +
                           M1->element[2][2] * M2->element[2][0];

      out->element[2][1] = M1->element[2][0] * M2->element[0][1] +
                           M1->element[2][1] * M2->element[1][1] +
                           M1->element[2][2] * M2->element[2][1];

      out->element[2][2] = M1->element[2][0] * M2->element[0][2] +
                           M1->element[2][1] * M2->element[1][2] +
                           M1->element[2][2] * M2->element[2][2];

      out->element[2][3] = M1->element[2][0] * M2->element[0][3] +
                           M1->element[2][1] * M2->element[1][3] +
                           M1->element[2][2] * M2->element[2][3] +
                           M1->element[2][3];
}

//############################################################################
//##                                                                        ##
//## Create a rotation matrix with specified axis and angle in radians      ##
//##                                                                        ##
//############################################################################

void TRANSFORM_compose_rotation(TRANSFORM *X, U32 axis, SINGLE angle)
{
   SINGLE _cos,_sin;

   *X = identity;

   _sin = sin(angle * MUL_DEG_TO_RAD);
   _cos = cos(angle * MUL_DEG_TO_RAD);

   switch (axis)
      {
      case PITCH:

         X->element[1][1] =  _cos; X->element[1][2] = -_sin;
         X->element[2][1] =  _sin; X->element[2][2] =  _cos;
         break;

      case YAW:

         X->element[0][0] =  _cos; X->element[0][2] =  _sin;
         X->element[2][0] = -_sin; X->element[2][2] =  _cos;
         break;

      case ROLL:

         X->element[0][0] =  _cos; X->element[0][1] = -_sin;
         X->element[1][0] =  _sin; X->element[1][1] =  _cos;
         break;
      }
}

//############################################################################
//##                                                                        ##
//## Transpose a matrix                                                     ##
//##                                                                        ##
//############################################################################

TRANSFORM *TRANSFORM_transpose(TRANSFORM *X)
{
   S32          i,j;
   static TRANSFORM transpose;

   for (i=0; i < 3; i++)
      {
      for (j=0; j < 3; j++)
         {
         transpose.element[i][j] = X->element[j][i];
         }

      transpose.element[i][3] = -X->element[i][3];
      }

   return &transpose;
}

//############################################################################
//##                                                                        ##
//## Invert a transformation matrix                                         ##
//##                                                                        ##
//############################################################################

TRANSFORM *TRANSFORM_invert(TRANSFORM *X)
{
   TRANSFORM        xlt;
   TRANSFORM        inverse;
   static TRANSFORM output;

   //
   // Transpose the matrix
   //

   inverse = *TRANSFORM_transpose(X);

   //
   // Create matrix with transposed translation element only (no rotation)
   //
   
   xlt = identity;

   xlt.element[0][3] = inverse.element[0][3];
   xlt.element[1][3] = inverse.element[1][3];
   xlt.element[2][3] = inverse.element[2][3];

   //
   // Zero translation element of original matrix
   //

   inverse.element[0][3] = 0;
   inverse.element[1][3] = 0;
   inverse.element[2][3] = 0;

   //
   // Multiply and return result
   //

   TRANSFORM_multiply(&inverse, &xlt, &output);

   return &output;
}

//############################################################################
//##                                                                        ##
//## Return the position (translation) component of a transform             ##
//##                                                                        ##
//############################################################################

void TRANSFORM_get_position(TRANSFORM *X, VECTOR *V)
{
   V->x = X->element[0][3];
   V->y = X->element[1][3];
   V->z = X->element[2][3];
}

//############################################################################
//##                                                                        ##
//## Set a transform's absolute position                                    ##
//##                                                                        ##
//############################################################################

void TRANSFORM_set_position(TRANSFORM *X, //)
                            SINGLE     x,
                            SINGLE     y,
                            SINGLE     z)
{
   X->element[0][3] = x;
   X->element[1][3] = y;
   X->element[2][3] = z;
}

//############################################################################
//##                                                                        ##
//## Adjust transform's relative position                                   ##
//##                                                                        ##
//############################################################################

void TRANSFORM_move_position(TRANSFORM *X, //)
                             SINGLE     dx,
                             SINGLE     dy,
                             SINGLE     dz)
{
   VECTOR in, out;

   in.x = dx;
   in.y = dy;
   in.z = dz;

   VECTOR_rotate(&in, &out, X);

   X->element[0][3] += out.x;
   X->element[1][3] += out.y;
   X->element[2][3] += out.z;
}

//############################################################################
//##                                                                        ##
//## Set transform's absolute orientation  (angles of incidence             ##
//## to X=0, Y=0, Z=0 planes)                                               ##
//##                                                                        ##
//############################################################################

void TRANSFORM_set_orientation(TRANSFORM *X, //)
                               SINGLE     pitch,
                               SINGLE     roll,
                               SINGLE     yaw)
{
   TRANSFORM work,out;

   //
   // Reset orientation element of camera transform, leaving position
   // vector untouched
   //

   X->element[0][0] = 1.0;
   X->element[1][0] = 0.0;
   X->element[2][0] = 0.0;

   X->element[0][1] = 0.0;
   X->element[1][1] = 1.0;
   X->element[2][1] = 0.0;

   X->element[0][2] = 0.0;
   X->element[1][2] = 0.0;
   X->element[2][2] = 1.0;
   
   //
   // Apply yaw, pitch, roll in order
   // 

   TRANSFORM_compose_rotation(&work,YAW,yaw);
   TRANSFORM_multiply(X, &work, &out);
   *X = out;

   TRANSFORM_compose_rotation(&work,PITCH,pitch);
   TRANSFORM_multiply(X, &work, &out);
   *X = out;

   TRANSFORM_compose_rotation(&work,ROLL,roll);
   TRANSFORM_multiply(X, &work, &out);
   *X = out;
}

//############################################################################
//##                                                                        ##
//## Load a saved transform from a DOS file                                 ##
//##                                                                        ##
//############################################################################

S32 TRANSFORM_load_from_file(TRANSFORM *X, //)
                             C8        *filename)
{                   
   *X = identity;

   FILE *in = fopen(filename,"rb");

   if (in != NULL)
      {
      S32 success = fread(X, sizeof(*X), 1, in);
      fclose(in);

      return success;
      }

   return 0;
}

