//############################################################################
//##                                                                        ##
//##  MESH.H: Object mesh handling (loading, instantiation, etc.)           ##
//##                                                                        ##
//############################################################################
//##                                                                        ##
//## Released to the public domain 12/15/99 by John Miles (jmiles@pop.net)  ##
//## No restrictions are placed on use or distribution of any of this code. ##
//##                                                                        ##
//############################################################################

#ifndef MESH_H
#define MESH_H

#include "gl/gl.h"
#include "fpmath.h"

#ifdef __cplusplus
extern "C" {
#endif

//
// Predefined limits
//

#define MAX_OBJECT_VERTICES  131072     // 128K vertices max. per object
#define MAX_OBJECT_FACES     65536      // 64K max faces per object

//
// Mesh material descriptor
//

typedef struct _MATERIAL
{
   U32         bitmap_width;       // Width/height of texture
   U32         bitmap_height;

   union                           // Stored copy of bitmap for OpenGL
      {  
      U32 *u32;                    // RGBA 
      U8  *u8;                     // A only
      } 
   txt;

   GLuint      txtID;              // OpenGL ID

   U32         valid;              // TRUE if there is a valid texture for this material
   U8         *palette;
}                 
MATERIAL;

//
// 2D vertex in texture space
//

typedef struct
{
   SINGLE u;
   SINGLE v;
}
TEXTURE_VERTEX;

//
// Runtime object mesh structure
//

typedef struct
{
   //
   // ASCII mesh name (for debugging)
   //

   C8 mesh_name[128];

   //
   // Array size limits for this MESH object (used during construction, 
   // expansion, etc.)
   //

   U32 material_limit;
   U32 normal_limit;
   U32 texture_vertex_limit;
   U32 object_vertex_limit;
   U32 vertex_reference_limit;
   U32 face_limit;
   U32 BSP_node_limit;

   //
   // List of bitmapped textures used in mesh
   //

   U32                material_cnt; 
   MATERIAL          *material_list;

   //
   // List of unique plane orientations (surface normals) for all faces 
   // (and BSP nodes) in mesh
   //
   // Surface normal (plane) equation: Ax+By+Cz-D=0, where A, B, and C
   // are coefficients that uniquely specify a plane's orientation, and D
   // specifies the position of the plane in object space (i.e., its position
   // along its infinite perpendicular axis)
   // 
   // All parallel faces in the mesh share the same A, B, and C coefficients, 
   // but only coplanar faces share the same D coefficient. This property 
   // is useful for rapid backface culling and BSP traversal, as it allows us
   // to re-use dot product calculations between all faces which are either
   // coplanar or parallel to each other.
   //

   U32     normal_cnt;          // # of unique surface normals 
   VECTOR *normal_ABC;          // Normal A, B, and C coefficients (Ax+By+Cz)

   //
   // List of 2D texture-space vertices in mesh
   //

   U32             texture_vertex_cnt;
   TEXTURE_VERTEX *texture_vertex_list;

   //
   // List of 3D object-space vertices in mesh
   //

   U32     object_vertex_cnt;          // # of unique object-space vertices
   VECTOR *object_vertex_list;      
   
   //
   // List of object-space and texture-space vertex chains for faces
   // These arrays are lists of indices to successive vertices in each face
   //

   U32     vertex_reference_cnt;       // # of vertex references

   U32    *object_vertex_chain;        // Ordered lists of vertex references
   U32    *texture_vertex_chain;       // for each face

   //
   // List of faces in mesh, defined as a closed list of vertices in object
   // (XYZ) and texture (UV) space, with an associated surface normal 
   // (plane orientation in object space) and D-coefficient (plane offset 
   // along its normal axis)
   //

   U32     face_cnt;                   // # of faces

   U32    *face_num_vertices;          // # of vertices in face
   U32    *face_vertices;              // Offset in obj/txt vertex chains

   U32    *face_material;              // Indices into material descriptor list
   U32    *face_normal;                // Indices into surface normal list
   SINGLE *face_D_coefficient;         // D-coefficients of normals

   SINGLE   *face_u0;                  // Texture U/V extents for faces
   SINGLE   *face_v0;
   SINGLE   *face_u1;              
   SINGLE   *face_v1;

   MATERIAL *face_lightmap;            // Light map materials for faces

   //
   // Optional BSP tree and associated data
   //
   // If BSP_node_cnt == 0, none of the other BSP-related fields are valid
   //
                              
   S32     BSP_node_cnt;               // # of BSP tree nodes (0 if no tree)
   U32     BSP_root_node;              // BSP root node index
   U32    *BSP_normal;                 // Indices into surface normal list
   SINGLE *BSP_D_coefficient;          // D-coefficients of plane normals
   S32    *BSP_num_node_faces;         // # of faces embedded in node's plane
   U32    *BSP_node_faces;             // Offset in node face chain
   S32    *BSP_front_child;            // Children of this node: node if +/0,
   S32    *BSP_back_child;             // leaf if - (-1=leaf 0, -2=1...)

   S32     node_face_reference_cnt;    // # of node face references
   U32    *node_face_chain;            // Ordered lists of embedded faces

   S32     BSP_leaf_cnt;               // # of BSP terminal (leaf) nodes
   S32    *BSP_num_leaf_faces;         // # of faces contained at each leaf
   U32    *BSP_leaf_faces;             // Offset in leaf face chain

   S32     leaf_face_reference_cnt;    // # of leaf face references
   U32    *leaf_face_chain;            // Ordered lists of included faces

   //
   // Optional leaf-to-leaf visibility matrix (assumes presence of valid
   // BSP tree)
   //
   // If visibility_list_size == 0, none of the other visibility-related
   // fields are valid
   //

   U32     visibility_list_size;       // Size of visibility list chain
   U8     *visibility_lists;           // Ordered leaf occlusion lists
   U32    *BSP_leaf_visibility_list;   // Offset to leaf's occlusion list
}
MESH;

//
// Prototypes
//

extern MESH *MESH_create(S32 n_materials,  
                         S32 n_faces,
                         S32 n_vertex_references,
                         S32 n_unique_obj_vertices,
                         S32 n_unique_txt_vertices,
                         S32 n_normals);

extern void MESH_destroy(MESH *mesh);

extern S32  MESH_open_face(MESH     *mesh,
                           MATERIAL *material,
                           VECTOR   *normal,
                           SINGLE    D_coefficient);

extern S32  MESH_add_vertex(MESH           *mesh,
                            TEXTURE_VERTEX *txt_vertex,
                            VECTOR         *obj_vertex);

extern void MESH_close_face(MESH           *mesh);

extern MESH *MESH_load_from_file(C8 *filename);

extern MESH *MESH_create_from_BSP(C8 *BSP_filename);

void MATERIAL_bind_texture(MATERIAL *material);

void MESH_invalidate_textures(MESH *mesh);

#ifdef __cplusplus
}
#endif
   
#endif
