//############################################################################
//##                                                                        ##
//##  openFLAME                                                             ##
//##                                                                        ##
//##  OBJECT.C: Rendering and manipulation of mesh objects                  ##
//##                                                                        ##
//##  Flat-model source compatible with IBM 32-bit ANSI C/C++               ##
//##                                                                        ##
//##  Version 1.00 of 16-Sep-95: Initial                                    ##
//##                                                                        ##
//##  Author: John Miles                                                    ##
//##                                                                        ##
//############################################################################
//##                                                                        ##
//## Released to the public domain 12/15/99 by John Miles (jmiles@pop.net)  ##
//## No restrictions are placed on use or distribution of any of this code. ##
//##                                                                        ##
//############################################################################

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include "mesh.h"
#include "object.h"
#include "typedefs.h"
#include "system.h"
#include "userio.h"

extern HWINDOW primary;
extern GLVTABLE *GL;

//
// Static variables used to accelerate BSP tree recursion
//

static U32    draw_list   [MAX_OBJECT_FACES];
static SINGLE dot_products[MAX_OBJECT_VERTICES];
static U32    object_render_cycle = 0;

static MESH     *mesh;
static OBJECT   *object;
static U32      *draw_index;
static SINGLE   *D;
static U32      *normal_index;

S32 wire_mode = 0;
S32 BSP_leaf;

//############################################################################
//##                                                                        ##
//## Return BSP leaf # for point in world space                             ##
//##                                                                        ##
//############################################################################

S32 BSP_classify_point(OBJECT *obj, //)
                       SINGLE  x,
                       SINGLE  y,
                       SINGLE  z)
{
   VECTOR      *normal;
   S32          cnt;
   TRANSFORM   *world_to_object;
   VECTOR       point_in_object_space;
   VECTOR       point_in_world_space;
   SINGLE      *dot;
   S32          node,child,where;

   //
   // Transform point from world to object space
   //

   mesh = obj->mesh;

   point_in_world_space.x = x;
   point_in_world_space.y = y;
   point_in_world_space.z = z;

   world_to_object = TRANSFORM_invert(&obj->to_world);

   VECTOR_rotate_translate(&point_in_world_space,
                           &point_in_object_space,
                            world_to_object);

   //
   // Pre-calculate each vector-normal dot product for later comparison with 
   // each face's plane
   //

   normal = mesh->normal_ABC;
   cnt    = mesh->normal_cnt;
   dot    = dot_products;

   while (--cnt >= 0)
      {
      *dot++   = normal->x * point_in_object_space.x +
                 normal->y * point_in_object_space.y +
                 normal->z * point_in_object_space.z;
      ++normal;
      }

   //
   // Filter point through BSP tree, if available
   // 

   if (!mesh->BSP_node_cnt)
      {
      return -1;
      }

   node = mesh->BSP_root_node;

   while (1)
      {
      if (dot_products[mesh->BSP_normal[node]] + mesh->BSP_D_coefficient[node] >= 0)
         {
         child = mesh->BSP_front_child[node];
         where = 1;
         }
      else
         {
         child = mesh->BSP_back_child[node];
         where = 0;
         }

      if (child <= 0)
         {
         break;
         }

      node = child;
      }

   //
   // Return current leaf #
   //

   return -(child+1);
}

//############################################################################
//##                                                                        ##
//## Recursively traverse BSP tree to generate ordered list of polygons     ##
//## that face the camera                                                   ##
//##                                                                        ##
//############################################################################

//
// Generate front-to-back (inverse painters') draw list, suitable for
// extent or span-based overdraw suppression
//

void BSP_traverse_FB(U32 node)
{
   static S32  i;
   static U32 *face;

   //
   // If this node is facing us, add any children in front of it that
   // are facing us to the list, then add this node, then add any
   // children behind it that are facing us to the list
   //
   // If this node is NOT facing us, the sense of "behind" and "front" are
   // reversed -- all of the node's "front" children are farther from our
   // POV, and must be traversed after the node's "back," or closer, 
   // children.  We do not add the current node to the list in this case; we
   // simply continue our traversal with the closer subtree after the
   // far traversal returns.
   //
   // The dot product comparison tells us two things in this routine: whether
   // or not to add the node to the visible-faces list, and the order in which
   // to traverse the node's left and right subtrees.
   //

   if (dot_products[mesh->BSP_normal[node]] + mesh->BSP_D_coefficient[node] >= 0)
      {
      //
      // Traverse all front children of this node
      //

      if (mesh->BSP_front_child[node] > 0)
         {
         BSP_traverse_FB(mesh->BSP_front_child[node]);
         }

      //
      // Add all visible coplanar faces at this node to draw list
      //

      face = &mesh->node_face_chain[mesh->BSP_node_faces[node]];

      for (i=0; i < mesh->BSP_num_node_faces[node]; i++)
         {
         if ((!object->visibility_valid)
              ||
             (object->face_marked_visible[*face] == object->face_mark_cycle))
            {
            if (dot_products[mesh->face_normal[*face]] + mesh->face_D_coefficient[*face] >= 0)
               {
               *draw_index++ = *face;
               }
            }

         face++;
         }

      //
      // Traverse all back children of this node
      //

      if (mesh->BSP_back_child[node] > 0)
         {
         BSP_traverse_FB(mesh->BSP_back_child[node]);
         }
      }
   else
      {
      if (mesh->BSP_back_child[node] > 0)
         {
         BSP_traverse_FB(mesh->BSP_back_child[node]);
         }

      //
      // Add all coplanar faces at this node to draw list
      //

      face = &mesh->node_face_chain[mesh->BSP_node_faces[node]];

      for (i=0; i < mesh->BSP_num_node_faces[node]; i++)
         {
         if ((!object->visibility_valid)
              ||
             (object->face_marked_visible[*face] == object->face_mark_cycle))
            {
            if (dot_products[mesh->face_normal[*face]] + mesh->face_D_coefficient[*face] >= 0)
               {
               *draw_index++ = *face;
               }
            }

         face++;
         }

      if (mesh->BSP_front_child[node] > 0)
         {
         BSP_traverse_FB(mesh->BSP_front_child[node]);
         }
      }
}

//
// Generate back-to-front (normal painters') draw list, suitable for
// non-Z-buffered rendering
//

void BSP_traverse_BF(U32 node)
{
   static S32  i;
   static U32 *face;

   //
   // If this node is facing us, add any children in front of it that
   // are facing us to the list, then add this node, then add any
   // children behind it that are facing us to the list
   //
   // If this node is NOT facing us, the sense of "behind" and "front" are
   // reversed -- all of the node's "front" children are farther from our
   // POV, and must be traversed after the node's "back," or closer, 
   // children.  We do not add the current node to the list in this case; we
   // simply continue our traversal with the closer subtree after the
   // far traversal returns.
   //
   // The dot product comparison tells us two things in this routine: whether
   // or not to add the node to the visible-faces list, and the order in which
   // to traverse the node's left and right subtrees.
   //

   if (dot_products[mesh->BSP_normal[node]] + mesh->BSP_D_coefficient[node] >= 0)
      {
      //
      // Traverse all back children of this node
      //

      if (mesh->BSP_back_child[node] > 0)
         {
         BSP_traverse_BF(mesh->BSP_back_child[node]);
         }

      //
      // Add all visible coplanar faces at this node to draw list
      //

      face = &mesh->node_face_chain[mesh->BSP_node_faces[node]];

      for (i=0; i < mesh->BSP_num_node_faces[node]; i++)
         {
         if ((!object->visibility_valid)
              ||
             (object->face_marked_visible[*face] == object->face_mark_cycle))
            {
            if (dot_products[mesh->face_normal[*face]] + mesh->face_D_coefficient[*face] >= 0)
               {
               *draw_index++ = *face;
               }
            }

         face++;
         }

      //
      // Traverse all front children of this node
      //

      if (mesh->BSP_front_child[node] > 0)
         {
         BSP_traverse_BF(mesh->BSP_front_child[node]);
         }
      }
   else
      {
      if (mesh->BSP_front_child[node] > 0)
         {
         BSP_traverse_BF(mesh->BSP_front_child[node]);
         }

      //
      // Add all coplanar faces at this node to draw list
      //

      face = &mesh->node_face_chain[mesh->BSP_node_faces[node]];

      for (i=0; i < mesh->BSP_num_node_faces[node]; i++)
         {
         if ((!object->visibility_valid)
              ||
             (object->face_marked_visible[*face] == object->face_mark_cycle))
            {
            if (dot_products[mesh->face_normal[*face]] + mesh->face_D_coefficient[*face] >= 0)
               {
               *draw_index++ = *face;
               }
            }

         face++;
         }

      if (mesh->BSP_back_child[node] > 0)
         {
         BSP_traverse_BF(mesh->BSP_back_child[node]);
         }
      }
}

//############################################################################
//##                                                                        ##
//## Destroy an object                                                      ##
//##                                                                        ##
//############################################################################

void OBJECT_destroy(OBJECT *object)
{
   if (object->face_marked_visible != NULL)
      {
      free(object->face_marked_visible);
      }

   free(object);
}

//############################################################################
//##                                                                        ##
//## Create an object                                                       ##
//##                                                                        ##
//############################################################################

OBJECT * OBJECT_create(MESH *mesh)
{                   
   OBJECT *object;

   //
   // Allocate object and zero all fields
   //

   object = (OBJECT *) malloc(sizeof(OBJECT));

   if (object == NULL)
      {
      alert_box("Error","Out of memory\n");
      return NULL;
      }

   memset(object, 0, sizeof(OBJECT));

   //
   // Associate MESH with object instance
   //

   object->mesh = mesh;

   //
   // Set up identity transform for object
   //

   object->to_world  = identity;

   //
   // If this object has precomputed visibility lists, allocate frame-coherent
   // list for currently-occupied leaf node
   // 

   object->current_leaf        = -1;
   object->face_mark_cycle     =  0;
   object->face_marked_visible =  NULL;
   object->visibility_valid    =  FALSE;

   if ((mesh->BSP_node_cnt         != 0) &&
       (mesh->visibility_list_size != 0))
      {
      object->face_marked_visible = (U32 *) 
               malloc(sizeof(U32) * mesh->face_cnt);

      if (object->face_marked_visible == NULL)
         {
         alert_box("Error","Out of memory!\n");
         return NULL;
         }

      //
      // Mark every face visible as of cycle 0
      //

      memset(object->face_marked_visible,
             0,
             sizeof(U32) * mesh->face_cnt);
      }

   return object;
}

//############################################################################
//##                                                                        ##
//## Render object into scene                                               ##
//##                                                                        ##
//############################################################################

void OBJECT_render (OBJECT    *target, //)
                    TRANSFORM *view_to_world)
{
   U8             *vis;
   S32             len,leaf;
   S32             draw_cnt;
   S32             face;
   S32             face_cnt;
   S32             f,i,j,set,n,offset;
   U32             n_vertices;
   U32            *tptr;
   U32            *vptr;
   U32            *f_index;
   SINGLE         *dot;
   TEXTURE_VERTEX *tv_list;
   TRANSFORM       object_to_view;
   TRANSFORM      *world_to_view;
   VECTOR         *vertex_list;

   //
   // Increment cycle count (used to invalidate transformed vertices, marked-
   // visible faces, etc.)
   // 

   ++object_render_cycle;

   //
   // Set up common local aliases for speed
   //

   object = target;
   mesh   = object->mesh;

   vertex_list    = mesh->object_vertex_list;
   tv_list        = mesh->texture_vertex_list;

   //
   // Obtain world-to-view transform by inverting the view-to-world transform
   // (i.e., the camera's position and orientation in world space)
   //

   world_to_view = TRANSFORM_invert(view_to_world);

   //
   // Concatenate world-to-view and object-to-world transforms to get 
   // object-to-view transform
   //

   TRANSFORM_multiply(world_to_view,
                     &object->to_world,
                     &object_to_view);

   //
   // Pre-calculate each vector-normal dot product for later comparison with 
   // each face's plane
   //

   BSP_leaf = BSP_classify_point(object,
                                 view_to_world->element[0][3],
                                 view_to_world->element[1][3],
                                 view_to_world->element[2][3]);

   // -------------------------------------------------
   // Determine potentially-visible-set (PVS) by means of visibility lists,
   // if available for this object
   // -------------------------------------------------

   if (mesh->BSP_node_cnt != 0)
      {
      //
      // If POV has moved to a new BSP leaf node...
      //

      if (BSP_leaf != object->current_leaf)
         {
         object->current_leaf = BSP_leaf;

         //
         // If BSP visibility lists are supported by mesh...
         //

         if (mesh->visibility_list_size != 0)
            {
            //
            // If a valid visibility list entry exists for the current leaf...
            //

            if ((object->current_leaf == 0)
                 ||
               ((offset = mesh->BSP_leaf_visibility_list[object->current_leaf]) < 0))
               {
               //
               // Object has visibility information, but not for this 
               // particular leaf, so set up to draw all polygons
               //

               object->visibility_valid = FALSE;
               }
            else
               {
               vis = mesh->visibility_lists + offset;

               len = ((mesh->BSP_leaf_cnt + 7) >> 3);
 
               set = 0;

               while (set < len)
                  {
                  n = *vis++;

                  if (!n)
                     {   
                     //
                     // Skip a given number of sets of 8 occluded leaf nodes
                     //

                     set += *vis++;
                     }
                  else
                     {
                     //
                     // At least one of 8 leaves in this set is visible -- 
                     // determine which leaf node(s) are potentially visible 
                     // and mark all of their polygons accordingly
                     //

                     for (i=0; i < 8; i++)
                        {
                        if (n & (1 << i))
                           {
                           leaf = ((set << 3) + i) + 1;

                           if (leaf >= mesh->BSP_leaf_cnt)
                              {
                              continue;
                              }

                           //
                           // Mark all surfaces in this leaf node as members 
                           // of the potentially-visible set being established
                           // this cycle
                           //

                           f_index = &mesh->leaf_face_chain[mesh->BSP_leaf_faces[leaf]];

                           for (j=0; j < mesh->BSP_num_leaf_faces[leaf]; j++)
                              {
                              object->face_marked_visible [ *f_index++ ] =
                                 object_render_cycle;
                              }
                           }
                        }

                     //
                     // Advance to the next set of 8 leaves
                     //

                     set++;
                     }
                  }

               //
               // Mark presence of valid visibility information in this object
               //
               // Log render cycle at which the current PVS became valid
               //

               object->visibility_valid = TRUE;
               object->face_mark_cycle  = object_render_cycle;
               }
            }
         }
      }

   // -------------------------------------------------
   // Build list of polygons that face the camera POV
   //
   // When possible, perform depth-first traversal of BSP tree to 
   // sort polygon list
   // -------------------------------------------------

   draw_index   = draw_list;
   dot          = dot_products;

   if (mesh->BSP_node_cnt)
      {
      //
      // BSP tree present -- create valid draw-order list
      //

      BSP_traverse_BF(mesh->BSP_root_node);
      }
   else
      {
      //
      // No BSP tree available -- add all camera-facing polygons to list
      //

      normal_index = mesh->face_normal;
      D            = mesh->face_D_coefficient;

      face_cnt = mesh->face_cnt;

      for (face=0; face < face_cnt; face++)
         {
         if (dot_products[*normal_index++] + *D++ >= 0)
            {
            *draw_index++ = face;
            }
         }
      }

   // -------------------------------------------------
   // Get final list of polygons to draw
   // -------------------------------------------------

   draw_cnt = ((U32) draw_index - (U32) draw_list) /
              sizeof(U32);

   // -------------------------------------------------
   // Set up object-to-view transform, transposing rows and
   // columns for OpenGL compliance
   // -------------------------------------------------

   GL->glMatrixMode(GL_MODELVIEW);

   GLfloat m[4][4];

   for (S32 mr=0; mr < 3; mr++)
      {
      for (S32 mc=0; mc < 4; mc++)
         {
         m[mc][mr] = object_to_view.element[mr][mc];
         }
      }

   m[0][3] = 0.0F;
   m[1][3] = 0.0F;
   m[2][3] = 0.0F;
   m[3][3] = 1.0F;

   GL->glLoadMatrixf((GLfloat *) m);

   // -------------------------------------------------
   // For all potentially-visible faces in object...
   // -------------------------------------------------

   for (f=0; f < draw_cnt; f++)
      {
      //
      // Process faces in predetermined order
      //

      face = draw_list[f];

      //
      // Set up local aliases for speed
      // 
      // vptr       = address of object vertex index sublist for this face
      // tptr       = address of texture vertex index sublist for this face
      // n_vertices = # of vertices in this face
      // 

      tptr = &mesh->texture_vertex_chain[mesh->face_vertices [face]];
      vptr = &mesh-> object_vertex_chain[mesh->face_vertices [face]];

      n_vertices = mesh->face_num_vertices[face];

      U32 *vp;
      U32 *tp;

      //
      // Disable Z-buffering
      //

      GL->glDisable(GL_DEPTH_TEST);

      // ------------------------------------
      // Draw texture map
      // ------------------------------------

      MATERIAL_bind_texture(&mesh->material_list[mesh->face_material[face]]);

      GL->glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

      GL->glDisable(GL_BLEND);

      GL->glBegin(GL_TRIANGLE_FAN);

      n  = n_vertices;
      vp = vptr;
      tp = tptr;

      while (n--)
         {
         VECTOR *vertex = &vertex_list[*vp++];

         SINGLE u = tv_list[*tp  ].u;
         SINGLE v = tv_list[*tp++].v;

         GL->glTexCoord2f (u, v);
         GL->glVertex3f   (vertex->x, vertex->y, vertex->z);
         }

      GL->glEnd();

      // ------------------------------------
      // Modulate with lightmap texture
      // ------------------------------------

      MATERIAL_bind_texture(&mesh->face_lightmap[face]);

      GL->glBlendFunc(GL_ONE_MINUS_SRC_ALPHA, GL_SRC_ALPHA);

      GL->glEnable(GL_BLEND);

      GL->glBegin(GL_TRIANGLE_FAN);

      n  = n_vertices;
      vp = vptr;
      tp = tptr;

      while (n--)
         {
         VECTOR *vertex = &vertex_list[*vp++];

         SINGLE u = tv_list[*tp  ].u;
         SINGLE v = tv_list[*tp++].v;

         //
         // Generate linearly-interpolated lightmap coordinates based
         // on texture vertices' position within polygon boundaries
         //

         SINGLE u0 = mesh->face_u0[face];
         SINGLE u1 = mesh->face_u1[face];
         SINGLE v0 = mesh->face_v0[face];
         SINGLE v1 = mesh->face_v1[face];

         u = (u-u0) / (u1-u0);
         v = (v-v0) / (v1-v0);

         GL->glTexCoord2f (u, v);
         GL->glVertex3f   (vertex->x, vertex->y, vertex->z);
         }

      GL->glEnd();

      GL->glDisable(GL_BLEND);

      // ------------------------------------
      // Draw optional wireframe around polygon
      // ------------------------------------

      if (wire_mode)
         {
         //
         // Trace wireframe outline around face
         //

         GL->glDisable(GL_TEXTURE_2D);

         GL->glBegin(GL_LINE_LOOP);

         n  = n_vertices;
         vp = vptr;

         while (n--)
            {
            VECTOR *vertex = &vertex_list[*vp++];

            GL->glColor3f(1.0F, 0.0F, 1.0F);
            GL->glVertex3d(vertex->x, vertex->y, vertex->z);
            }

         GL->glEnd();
         }

      // ------------------------------------
      // End processing for this face
      // ------------------------------------
      }
} 
